# The MIT License (MIT)
# Copyright (c) 2021 by the ESA CCI Toolbox development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from xcube.core.timecoord import get_timestamps_from_string

from xcube_cci.constants import MONTHS


def get_timestrings_from_string(string: str) -> (str, str):
    first_time, second_time = get_timestamps_from_string(string)
    if first_time:
        first_time = first_time.isoformat()
    if second_time:
        second_time = second_time.isoformat()
    if not first_time:
        for i, month in enumerate(MONTHS):
            if month in string:
                first_time = i + 1
                break
    return first_time, second_time



# The MIT License (MIT)
# Copyright (c) 2024 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Any, Dict, Hashable, Mapping, Optional, Tuple

import dask.array as da
import numpy as np
import xarray
from xcube.core.store import (DataDescriptor, DataType, DataTypeLike,
                              VariableDescriptor)
from xcube.util.assertions import assert_true
from xcube.util.jsonschema import JsonIntegerSchema, JsonObjectSchema

from .constants import LOG


class VectorDataCube(xarray.Dataset):
    """A wrapper class around an xarray Dataset to keep it separate."""
    __slots__ = ()

VECTOR_DATA_CUBE_TYPE = DataType(
    VectorDataCube,
    ['vectordatacube', 'xarray.Dataset']
)

DataType.register_data_type(VECTOR_DATA_CUBE_TYPE)


def _attrs_to_json(attrs: Mapping[Hashable, Any]) \
        -> Optional[Dict[str, Any]]:
    new_attrs: Dict[str, Any] = {}
    for k, v in attrs.items():
        if isinstance(v, np.ndarray):
            v = v.tolist()
        elif isinstance(v, da.Array):
            v = np.array(v).tolist()
        if isinstance(v, float) and np.isnan(v):
            v = None
        new_attrs[str(k)] = v
    return new_attrs


class VectorDataCubeDescriptor(DataDescriptor):
    """
    A descriptor for an N-dimensional vector data cube whose spatial properties are
    designated by a geometry cooordinate. The cube is represented by an xarray.Dataset.
    Comprises a description of the data variables contained in the dataset.

    Regarding *time_range* and *time_period* parameters, please refer to
    https://github.com/dcs4cop/xcube/blob/master/docs/source/storeconv.md#date-time-and-duration-specifications

    :param data_id: An identifier for the data
    :param data_type: The data type of the data described
    :param crs: A coordinate reference system identifier,
        as an EPSG, PROJ or WKT string
    :param bbox: A bounding box of the data
    :param time_range: Start and end time delimiting this data's
        temporal extent
    :param time_period: The data's periodicity
        if it is evenly temporally resolved
    :param dims: A mapping of the dataset's dimensions to their sizes
    :param coords: mapping of the dataset's data coordinate names
        to instances of :class:VariableDescriptor
    :param data_vars: A mapping of the dataset's variable names
        to instances of :class:VariableDescriptor
    :param attrs: A mapping containing arbitrary attributes of the dataset
    :param open_params_schema: A JSON schema describing the parameters
        that may be used to open this data
    """

    def __init__(self,
                 data_id: str,
                 *,
                 data_type: DataTypeLike = VECTOR_DATA_CUBE_TYPE,
                 crs: str = None,
                 bbox: Tuple[float, float, float, float] = None,
                 time_range: Tuple[Optional[str], Optional[str]] = None,
                 time_period: str = None,
                 spatial_res: float = None,
                 dims: Mapping[str, int] = None,
                 coords: Mapping[str, 'VariableDescriptor'] = None,
                 data_vars: Mapping[str, 'VariableDescriptor'] = None,
                 attrs: Mapping[Hashable, any] = None,
                 open_params_schema: JsonObjectSchema = None,
                 **additional_properties):
        super().__init__(data_id=data_id,
                         data_type=data_type,
                         crs=crs,
                         bbox=bbox,
                         time_range=time_range,
                         time_period=time_period,
                         open_params_schema=open_params_schema)
        assert_true(VECTOR_DATA_CUBE_TYPE.is_super_type_of(data_type),
                    f'illegal data_type,'
                    f' must be compatible with {VECTOR_DATA_CUBE_TYPE!r}')
        if additional_properties:
            LOG.info(f'Additional properties received;'
                     f' will be ignored: {additional_properties}')
        self.dims = dict(dims) if dims else None
        self.spatial_res = spatial_res
        self.coords = coords if coords else None
        self.data_vars = data_vars if data_vars else None
        self.attrs = _attrs_to_json(attrs) if attrs else None

    @classmethod
    def get_schema(cls) -> JsonObjectSchema:
        schema = super().get_schema()
        schema.properties.update(
            dims=JsonObjectSchema(
                additional_properties=JsonIntegerSchema(minimum=0)
            ),
            coords=JsonObjectSchema(
                additional_properties=VariableDescriptor.get_schema()
            ),
            data_vars=JsonObjectSchema(
                additional_properties=VariableDescriptor.get_schema()
            ),
            attrs=JsonObjectSchema(
                additional_properties=True
            ),
        )
        schema.required = ['data_id', 'data_type']
        schema.additional_properties = False
        schema.factory = cls
        return schema



# The MIT License (MIT)
# Copyright (c) 2026 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import asyncio
import random
import threading
from typing import Optional

import aiohttp

from .constants import (DEFAULT_NUM_RETRIES, DEFAULT_RETRY_BACKOFF_BASE,
                        DEFAULT_RETRY_BACKOFF_MAX, LOG)


class SessionExecutor:

    def __init__(
            self,
            user_agent: str = None,
            enable_warnings: bool = False,
            num_retries: int = DEFAULT_NUM_RETRIES,
            retry_backoff_max: int = DEFAULT_RETRY_BACKOFF_MAX,
            retry_backoff_base: float = DEFAULT_RETRY_BACKOFF_BASE,
    ):
        self._headers = {'User-Agent': user_agent} if user_agent else None
        self._enable_warnings = enable_warnings
        self._num_retries = num_retries
        self._retry_backoff_max = retry_backoff_max
        self._retry_backoff_base = retry_backoff_base
        self._executor_loop = None
        self._executor_session = None
        self._executor_thread = None
        self._loop_lock = threading.Lock()

    def _ensure_executor_loop(self):
        if self._executor_loop is not None:
            return

        with self._loop_lock:
            if self._executor_loop is not None:
                return

            loop_created = threading.Event()

            def loop_runner():
                self._executor_loop = asyncio.new_event_loop()
                asyncio.set_event_loop(self._executor_loop)
                timeout = aiohttp.ClientTimeout(total=300, connect=30, sock_connect=30, sock_read=120)
                connector = aiohttp.TCPConnector(limit=50, loop=self._executor_loop, force_close=True)
                self._executor_session = aiohttp.ClientSession(
                    connector=connector,
                    headers=self._headers,
                    trust_env=True,
                    timeout=timeout
                )
                loop_created.set()
                self._executor_loop.run_forever()

            self._executor_thread = threading.Thread(
                target=loop_runner,
                daemon=True
            )
            self._executor_thread.start()

            loop_created.wait()

    def run_with_session(self, async_function, *params):
        self._ensure_executor_loop()

        async def _run_with_session_executor(e_function, *e_params):
            return await e_function(self._executor_session, *e_params)

        future = asyncio.run_coroutine_threadsafe(
            _run_with_session_executor(async_function, *params),
            self._executor_loop
        )
        return future.result()

    def get_response_content(self, url: str) -> Optional[bytes]:
        return self.run_with_session(self.get_response_content_from_session, url)

    async def get_response_content_from_session(
            self, session: aiohttp.ClientSession, url: str
    ) -> Optional[bytes]:
        num_retries = self._num_retries
        retry_backoff_max = self._retry_backoff_max
        retry_backoff_base = self._retry_backoff_base
        for i in range(num_retries):
            retry_min = 100
            try:
                async with session.get(url) as resp:
                    retry_min = int(resp.headers.get('Retry-After', '100'))
                    if resp.status == 200:
                        return await resp.read()
                    elif 500 <= resp.status < 600:
                        error_message = f"Error {resp.status}: Cannot access url."
                        if self._enable_warnings:
                            LOG.warning(error_message)
                        return None
                    elif resp.status == 429:
                        error_message = "Error 429: Too Many Requests."
                    else:
                        return None
            except (
                    aiohttp.ClientConnectionError,
                    aiohttp.ServerDisconnectedError,
                    aiohttp.ClientPayloadError,
                    asyncio.TimeoutError,
            ) as e:
                error_message = str(e)
            retry_backoff = random.random() * retry_backoff_max
            retry_total = retry_min + retry_backoff
            if self._enable_warnings:
                LOG.info(
                    f"{error_message} - attempt {i + 1}/{num_retries}, "
                    f"retrying in {retry_total / 1000:.2f} s"
                )
            await asyncio.sleep(retry_total / 1000)
            retry_backoff_max *= retry_backoff_base

        return None



import os
from unittest import skipIf
from unittest import TestCase

import pandas as pd

from xcube_cci.cciodp import CciOdp
from xcube_cci.dataframeaccess import DataFrameAccessor

GHG_DS_ID = "esacci.GHG.satellite-orbit-frequency.L2.CH4.SCIAMACHY.Envisat.IMAP.v7-2.r1"

class DataFrameAccessTest(TestCase):

    def setUp(self) -> None:
        ccicdc = CciOdp(data_type='geodataframe')
        self._dfa = DataFrameAccessor(ccicdc, GHG_DS_ID, {})

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_geodataframe(self):
        gdf = self._dfa.get_geodataframe()
        self.assertIsNotNone(gdf)

    @skipIf(os.environ.get('XCUBE_CCI_DISABLE_WEB_TESTS', '1') == '1',
            'XCUBE_CCI_DISABLE_WEB_TESTS = 1')
    def test_get_geodataframe_for_dataset(self):
        gdf = self._dfa._get_features_from_cci_cdc(
            GHG_DS_ID,
            (pd.Timestamp('2003-01-08 00:00:00'), pd.Timestamp('2003-01-08 23:59:59'))
        )
        self.assertIsNotNone(gdf)



# The MIT License (MIT)
# Copyright (c) 2025 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List

import lxml.etree as etree

from .constants import CCI_ODD_URL
from .sessionexecutor import SessionExecutor

ODD_NS = {
    'os': 'http://a9.com/-/spec/opensearch/1.1/',
    'param': 'http://a9.com/-/spec/opensearch/extensions/parameters/1.0/'
}

def _get_from_param_elem(param_elem: etree.Element):
    options = param_elem.findall('param:Option', namespaces=ODD_NS)
    if not options:
        return None
    if len(options) == 1:
        return options[0].get('value'), \
               int(options[0].get('label').split('(')[-1][:-1])
    return [(option.get('value'), int(option.get('label').split('(')[-1][:-1]))
            for option in options]


class OdpConnector:

    def __init__(
            self,
            user_agent: str,
            endpoint_description_url: str = CCI_ODD_URL
    ):
        self._session_executor = SessionExecutor(user_agent=user_agent)
        self._endpoint_description_url = endpoint_description_url

    def get_drs_ids(self) -> List[str]:
        meta_info_dict = self.extract_metadata_from_odd_url(self._endpoint_description_url)
        return meta_info_dict.get('drs_ids', [])

    def extract_metadata_from_odd_url(self, odd_url: str = None) -> dict:
        if not odd_url:
            return {}
        resp_content = self._session_executor.get_response_content(odd_url)
        if not resp_content:
            return {}
        return self.extract_metadata_from_odd(etree.XML(resp_content))

    @staticmethod
    def extract_metadata_from_odd(odd_xml: etree.XML) -> dict:
        metadata = {'num_files': {}}
        metadata_names = {
            'ecv': (['ecv', 'ecvs'], False),
            'frequency': (['time_frequency', 'time_frequencies'], False),
            'institute': (['institute', 'institutes'], False),
            'processingLevel': (['processing_level', 'processing_levels'], False),
            'productString': (['product_string', 'product_strings'], False),
            'productVersion': (['product_version', 'product_versions'], False),
            'dataType': (['data_type', 'data_types'], False),
            'sensor': (['sensor_id', 'sensor_ids'], False),
            'platform': (['platform_id', 'platform_ids'], False),
            'fileFormat': (['file_format', 'file_formats'], False),
            'drsId': (['drs_id', 'drs_ids'], True)
        }
        for param_elem in odd_xml.findall('os:Url/param:Parameter',
                                          namespaces=ODD_NS):
            if param_elem.attrib['name'] in metadata_names:
                element_names, add_to_num_files = \
                    metadata_names[param_elem.attrib['name']]
                param_content = _get_from_param_elem(param_elem)
                if param_content:
                    if type(param_content) is tuple:
                        metadata[element_names[0]] = param_content[0]
                        if add_to_num_files:
                            metadata['num_files'][param_content[0]] = \
                                param_content[1]
                    else:
                        names = []
                        for name, num_files in param_content:
                            names.append(name)
                            if add_to_num_files:
                                metadata['num_files'][name] = num_files
                        metadata[element_names[1]] = names
        return metadata


# The MIT License (MIT)
# Copyright (c) 2021 by the xcube development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import os
from typing import Any, Container, Tuple, Union

from xcube.core.store import DataStoreError, DataTypeLike, get_data_store_class
from xcube.util.jsonschema import JsonObjectSchema

from.constants import ZARR_LOCATION

if os.environ.get(ZARR_LOCATION, 'CEDA') == "OTC":
    CCI_ZARR_STORE_ENDPOINT = 'https://ceda-backup-data.obs.eu-nl.otc.t-systems.com/zarr'
    DATA_IDS_FILE_PATH = f'{CCI_ZARR_STORE_ENDPOINT}/data_ids.json'
    CCI_ZARR_STORE_PARAMS = dict(
        root=CCI_ZARR_STORE_ENDPOINT
    )
    DataStore = get_data_store_class('https')
else:
    CCI_ZARR_STORE_BUCKET_NAME = 'esacci'
    CCI_ZARR_STORE_ENDPOINT = 'https://cci-ke-o.s3-ext.jc.rl.ac.uk:443/'
    DATA_IDS_FILE_PATH = f'{CCI_ZARR_STORE_BUCKET_NAME}/data_ids.json'
    CCI_ZARR_STORE_PARAMS = dict(
        root=CCI_ZARR_STORE_BUCKET_NAME,
        storage_options=dict(
            anon=True,
            client_kwargs=dict(
                endpoint_url=CCI_ZARR_STORE_ENDPOINT,
            )
        )
    )
    DataStore = get_data_store_class('s3')


class CciZarrDataStore(DataStore):

    def __init__(self):
        super().__init__(**CCI_ZARR_STORE_PARAMS)

    @property
    def root(self) -> str:
        if self._root is None:
            with self._lock:
                root = self._raw_root
                self._root = root
        return self._root

    @classmethod
    def get_data_store_params_schema(cls) -> JsonObjectSchema:
        return JsonObjectSchema(additional_properties=False)

    def get_data_ids(self,
                     data_type: DataTypeLike = None,
                     include_attrs: Container[str] | bool = False,) -> \
            Union[list[str], list[tuple[str, dict[str, Any]]]]:
        # TODO: do not ignore names in include_attrs
        if self.fs.exists(DATA_IDS_FILE_PATH):
            return_tuples = include_attrs and include_attrs is not None
            with self.fs.open(DATA_IDS_FILE_PATH) as f:
                ids = json.load(f)
                for id in ids:
                    yield (id, {}) if return_tuples else id
        else:
            yield from super().get_data_ids(
                data_type=data_type,
                include_attrs=include_attrs
            )

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def get_data_writer_ids(self, data_type: DataTypeLike = None) -> \
            Tuple[str, ...]:
        return ()

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def get_write_data_params_schema(self, **kwargs) -> \
            JsonObjectSchema:
        return JsonObjectSchema(additional_properties=False)

    def write_data(self, *args, **kwargs) -> str:
        raise DataStoreError('The CciZarrDataStore is read-only.')

    # noinspection PyUnusedLocal,PyMethodMayBeStatic
    def get_delete_data_params_schema(self, **kwargs) -> \
            JsonObjectSchema:
        return JsonObjectSchema(additional_properties=False)

    def delete_data(self, *args):
        raise DataStoreError('The CciZarrDataStore is read-only.')



# The MIT License (MIT)
# Copyright (c) 2020 by the ESA CCI Toolbox development team and contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import json
import logging
import os
import os.path
import shutil
import time
import traceback
from typing import List, Optional

import click
import xarray as xr
import zarr.storage

from xcube_cci.cciodp import CciOdp
from xcube_cci.chunkstore import CciChunkStore

DEFAULT_OUTPUT = 'open-report'


@click.command()
@click.option('--output', 'output_dir',
              metavar='OUTPUT_DIR',
              default=DEFAULT_OUTPUT,
              help=f'Output directory. Defaults to "{DEFAULT_OUTPUT}".')
@click.option('--force',
              is_flag=True,
              help='Force deletion of an existing OUTPUT_DIR.')
@click.option('--cache', 'cache_size',
              metavar='CACHE_SIZE',
              help='Cache size, e.g. "100M" or "2G". If given, an in-memory LRU cache will be used.')
@click.option('--observe',
              is_flag=True,
              help='Dump data requests. '
                   'Note, turning this flag on will slightly increase measured durations.')
@click.argument('dataset_id', nargs=-1, required=False)
def gen_report(output_dir: str,
               force: bool,
               cache_size: Optional[str],
               observe: bool,
               dataset_id: List[str]):
    """
    Opens CCI ODP datasets and generates a report in OUTPUT_DIR from opening a dataset.

    If DATASET_ID are omitted, all ODP datasets are opened. Otherwise, only the given
    datasets will be opened. A DATASET_ID may be just a part of the actual dataset
    identifier to be selected, e.g. "CLOUD" or "5-day".

    For each selected dataset, the following reports are generated after opening it:

    \b
    * ${OUTPUT_DIR}/SUCCESS@${DATASET_ID}.json - on success only;
    * ${OUTPUT_DIR}/ERROR@${DATASET_ID}.json - on error only;
    * ${OUTPUT_DIR}/ERROR@${DATASET_ID}.txt - on error only, contains exception traceback.

    If the "--observe" flag is used, an additional report is generated comprising each
    individual data request made:

    * ${OUTPUT_DIR}/OBSERVED@${DATASET_ID}.txt - in any case.

    The program will exit with an error if OUTPUT_DIR exists. To force deletion of
    an existing OUTPUT_DIR, use the "--force" flag.

    """
    logging.basicConfig(format='%(asctime)s %(levelname)s: %(message)s',
                        level=logging.DEBUG,
                        datefmt='%Y-%m-%d %H:%M:%S')
    if os.path.isdir(output_dir):
        if force:
            shutil.rmtree(output_dir)
        else:
            raise click.ClickException(f'Output directory "{output_dir}" already exists.')
    if not os.path.isdir(output_dir):
        os.makedirs(output_dir)

    cache_size = parse_cache_size(cache_size)

    odp = CciOdp()

    selected_ds_ids = get_selected_dataset_ids(odp, dataset_id)

    logging.info(f'Running tests for {len(selected_ds_ids)} datasets...')

    id_to_ts = dict()

    for i, ds_id in enumerate(selected_ds_ids):
        data_type = ''
        observer, obs_fp = new_observer(output_dir, ds_id) if observe else (None, None)

        logging.info(f'Attempting to open #{i+1} of {len(selected_ds_ids)}: {ds_id}')

        t0 = time.perf_counter()

        try:
            store = CciChunkStore(odp, ds_id, observer=observer, trace_store_calls=True)
        except Exception as e:
            report_error(output_dir, ds_id, t0, 'CciChunkStore()', e, data_type)
            continue

        if cache_size > 0:
            store = zarr.storage.LRUStoreCache(store, max_size=cache_size)

        try:
            logging.info('Attempting to open zarr ...')
            ds = xr.open_zarr(store)
            data_type = 'dataset'
            report_success(output_dir, ds_id, t0, ds, data_type)
        except Exception as e:
            report_error(output_dir, ds_id, t0, 'xr.open_zarr()', e, data_type)

        id_to_ts[ds_id] = data_type

        if obs_fp is not None:
            obs_fp.close()

    report_type_specifiers(output_dir, ts_dict=id_to_ts)

    logging.info('Tests completed.')


def get_selected_dataset_ids(odp, ds_id_parts):
    all_ds_ids = odp.dataset_names
    if ds_id_parts:
        ds_id_parts_lc = [ds_id_part.lower() for ds_id_part in ds_id_parts]
        selected_ds_ids = set()
        for ds_id in all_ds_ids:
            ds_id_lc = ds_id.lower()
            if any(map(lambda ds_id_part: ds_id_part in ds_id_lc, ds_id_parts_lc)):
                selected_ds_ids.add(ds_id)
    else:
        selected_ds_ids = all_ds_ids
    selected_ds_ids = sorted(selected_ds_ids)
    return selected_ds_ids


def new_observer(output_dir, ds_id):
    path = os.path.join(output_dir, f'OBSERVED@{ds_id}.txt')
    fp = open(path, 'a')

    def observer(var_name: str = None, chunk_index=None, time_range=None, duration=None, exception=None):
        tr = f'{time_range[0]}/{time_range[1]}'
        data = f'{var_name};{chunk_index};{tr};{format_millis(duration)}'
        status = f'{type(exception).__name__}({exception})' if exception else 'OK'
        fp.write(f'{status};{data}\n')
        fp.flush()

    return observer, fp


def report_success(output_dir: str, ds_id: str, t0: float, ds: xr.Dataset, type_specifier: str):
    obj = ds_to_dict(ds_id, time.perf_counter() - t0, ds, type_specifier)
    path = os.path.join(output_dir, f'SUCCESS@{ds_id}.json')
    try:
        with open(path, 'w') as fp:
            json.dump(obj, fp, indent=2)
        logging.info(f'{ds_id} took {format_millis(obj["duration"])} seconds')
    except Exception as e:
        if os.path.isfile(path):
            os.remove(path)
        report_error(output_dir, ds_id, time.perf_counter() - t0, 'to_json()', e)


def report_error(output_dir: str, ds_id: str, t0: float, stage: str, e: Exception, type_specifier: str):
    obj = error_to_dict(ds_id, stage, time.perf_counter() - t0, e, type_specifier)
    json_path = os.path.join(output_dir, f'ERROR@{ds_id}.json')
    text_path = os.path.join(output_dir, f'ERROR@{ds_id}.txt')
    with open(json_path, 'w') as fp:
        json.dump(obj, fp, indent=2)
    with open(text_path, 'w') as fp:
        traceback.print_exc(file=fp)
    logging.error(f'{stage}: {ds_id} took {format_millis(obj["duration"])}: {e}')


def report_type_specifiers(output_dir: str, ts_dict: dict):
    json_path = os.path.join(output_dir, 'type_specifiers.json')
    with open(json_path, 'w') as fp:
        json.dump(ts_dict, fp, indent=2)


def ds_to_dict(ds_id: str, duration: float, ds: xr.Dataset, type_specifier: str) -> dict:
    return dict(ds_id=ds_id,
                status='ok',
                type_specifier=type_specifier,
                duration=duration,
                dataset=dict(sizes=dict(ds.sizes),
                             coord_vars=vars_to_list(ds.coords),
                             data_vars=vars_to_list(ds.data_vars),
                             attrs=dict(ds.attrs)))


def error_to_dict(ds_id: str, stage: str, duration: float, e: Exception, type_specifier: str) -> dict:
    return dict(ds_id=ds_id,
                typeSpecifier=type_specifier,
                status='error',
                duration=duration,
                error=dict(type=str(type(e)),
                           stage=stage,
                           message=f'{e}'))


def vars_to_list(vars) -> list:
    return [var_to_dict(k, v) for k, v in vars.items()]


def var_to_dict(name: str, var: xr.DataArray) -> dict:
    return dict(name=name,
                dtype=str(var.dtype),
                dims=list(var.dims),
                attrs=dict(var.attrs))


def parse_cache_size(cache_size: Optional[str]) -> int:
    if not cache_size:
        return 0
    cache_size_lc = cache_size.lower()

    def _parse_cache_size():
        for suffix, factor in (
                ('kib', 1024), ('mib', 1024 ** 2), ('gib', 1024 ** 3), ('tib', 1024 ** 4),
                ('b', 1), ('k', 1000), ('m', 1000 ** 2), ('g', 1000 ** 3), ('t', 1000 ** 4)):
            if cache_size_lc.endswith(suffix):
                return int(cache_size[0:-len(suffix)]) * factor
        return int(cache_size)

    try:
        size = _parse_cache_size()
        if size < 0:
            raise ValueError()
        return size
    except ValueError:
        raise click.ClickException(f'Invalid cache size: "{cache_size}"')


def format_millis(time: float) -> str:
    return f'{round(time * 1000)} ms'


if __name__ == '__main__':
    gen_report()



# The MIT License (MIT)
# Copyright (c) 2023 ESA Climate Change Initiative
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from datetime import datetime
from typing import Any, List, Mapping, Tuple, Union

import pandas as pd
from dateutil.relativedelta import relativedelta

from xcube_cci.cciodp import CciOdp
from xcube_cci.constants import MONTHS

from .constants import TIMESTAMP_FORMAT


def extract_time_range_as_strings(
        time_range: Union[Tuple, List]
) -> (str, str):
    if isinstance(time_range, tuple):
        time_start, time_end = time_range
    else:
        time_start = time_range[0]
        time_end = time_range[1]
    return \
        extract_time_as_string(time_start), \
        extract_time_as_string(time_end)


def extract_time_as_string(
        time_value: Union[pd.Timestamp, str]
) -> str:
    if isinstance(time_value, str):
        time_value = pd.to_datetime(time_value, utc=True)
    return time_value.tz_localize(None).isoformat()


class TimeRangeGetter:

    def __init__(self, cci_cdc: CciOdp, metadata: dict):
        self._cci_cdc = cci_cdc
        self._metadata = metadata

    def get_time_ranges(self, dataset_id: str,
                        params: Mapping[str, Any]) -> List[Tuple]:
        start_time, end_time, iso_start_time, iso_end_time = \
            self._extract_time_range_as_datetime(
                params.get('time_range',
                           self.get_default_time_range(dataset_id)))
        ecv = dataset_id.split('.')[1]
        time_period = dataset_id.split('.')[2]
        if time_period == 'day':
            start_time = datetime(year=start_time.year, month=start_time.month,
                                  day=start_time.day)
            end_time = datetime(year=end_time.year, month=end_time.month,
                                day=end_time.day,
                                hour=23, minute=59, second=59)
            delta = relativedelta(days=1)
        elif time_period == 'month' or time_period == 'mon':
            start_time = datetime(year=start_time.year, month=start_time.month,
                                  day=1)
            end_time = datetime(year=end_time.year, month=end_time.month, day=1)
            delta = relativedelta(months=1)
            end_time += delta
        elif time_period == 'year' or time_period == 'yr' and ecv != "BIOMASS":
            start_time = datetime(year=start_time.year, month=1, day=1)
            end_time = datetime(year=end_time.year, month=12, day=31)
            delta = relativedelta(years=1)
            if dataset_id.endswith("yr"):
                num_years = int(dataset_id[-3:-2])
                delta_years = relativedelta(years=num_years) - relativedelta(days=1)
                request_time_ranges = []
                this = start_time
                after = this + delta_years
                while after <= end_time:
                    pd_this = pd.Timestamp(datetime.strftime(this, TIMESTAMP_FORMAT))
                    pd_next = pd.Timestamp(datetime.strftime(after, TIMESTAMP_FORMAT))
                    request_time_ranges.append((pd_this, pd_next))
                    this = this + delta
                    after = after + delta
                return request_time_ranges
        elif time_period == 'climatology':
            return [(i + 1, i + 1) for i, month in enumerate(MONTHS)]
        elif time_period == "5-days" and ecv == "VEGETATION":
            return self.extract_5days_vegetation_time_range(dataset_id, iso_start_time, iso_end_time)
        else:
            end_time = end_time.replace(hour=23, minute=59, second=59)
            end_time_str = datetime.strftime(end_time, TIMESTAMP_FORMAT)
            iso_end_time = extract_time_as_string(end_time_str)
            request_time_ranges = self._cci_cdc.get_time_ranges_from_data(
                dataset_id, iso_start_time, iso_end_time
            )
            return request_time_ranges
        drs_end = dataset_id.split('.')[9]
        if drs_end == "greenland_gmb_mass_trends":
            return self.extract_greenland_gmb_time_series(
                dataset_id, iso_start_time, end_time, delta
            )
        request_time_ranges = []
        this = start_time
        while this < end_time:
            after = this + delta
            pd_this = pd.Timestamp(datetime.strftime(this, TIMESTAMP_FORMAT))
            pd_next = pd.Timestamp(datetime.strftime(after, TIMESTAMP_FORMAT))
            request_time_ranges.append((pd_this, pd_next))
            this = after
        return request_time_ranges

    def extract_5days_vegetation_time_range(self, dataset_id: str, start_time: datetime, end_time: datetime):
        file_time_ranges = self._cci_cdc.get_time_ranges_from_data(dataset_id, start_time, end_time)
        five_days = pd.Timedelta(days=5)
        one_day = pd.Timedelta(days=1, seconds=-1)
        actual_time_ranges = []
        for file_time_range in file_time_ranges:
            start_time = file_time_range[0]
            time = pd.Timestamp(year=start_time.year, month=1, day=1)
            end_time = pd.Timestamp(year=start_time.year, month=12, day=30)
            while time < end_time:
                actual_time_ranges.append((time, time + one_day))
                time += five_days
        return actual_time_ranges

    def extract_greenland_gmb_time_series(
            self, dataset_id: str, iso_start_time: str,
            end_time: datetime, delta: pd.Timedelta
    )  -> List[Tuple]:
        end_time = end_time.replace(hour=23, minute=59, second=59)
        end_time_str = datetime.strftime(end_time, TIMESTAMP_FORMAT)
        iso_end_time = extract_time_as_string(end_time_str)
        request_time_ranges = self._cci_cdc.get_time_ranges_from_data(
            dataset_id, iso_start_time, iso_end_time
        )
        start_time = request_time_ranges[0][0]
        end_time = request_time_ranges[-1][1]
        num_years = end_time.year - start_time.year + 1
        span_delta = relativedelta(
            years=num_years - self._metadata["time_chunking"]
        )
        adjusted_time_ranges = []
        span_start_time = start_time
        span_end_time = start_time + span_delta
        span_end_time = span_end_time.replace(hour=23, minute=59, second=59)
        while span_end_time <= end_time:
            pd_this = pd.Timestamp(datetime.strftime(
                span_start_time, TIMESTAMP_FORMAT)
            )
            pd_next = pd.Timestamp(datetime.strftime(
                span_end_time, TIMESTAMP_FORMAT)
            )
            adjusted_time_ranges.append((pd_this, pd_next))
            span_start_time += delta
            span_end_time += delta
        return adjusted_time_ranges

    @staticmethod
    def _extract_time_range_as_datetime(
            time_range: Union[Tuple, List]
    ) -> (datetime, datetime, str, str):
        iso_start_time, iso_end_time = extract_time_range_as_strings(
            time_range)
        start_time = datetime.strptime(iso_start_time, TIMESTAMP_FORMAT)
        end_time = datetime.strptime(iso_end_time, TIMESTAMP_FORMAT)
        return start_time, end_time, iso_start_time, iso_end_time

    def get_default_time_range(self, ds_id: str):
        temporal_start = self._metadata.get('temporal_coverage_start', None)
        temporal_end = self._metadata.get('temporal_coverage_end', None)
        if not temporal_start or not temporal_end:
            time_ranges = self._cci_cdc.get_time_ranges_from_data(ds_id)
            if not temporal_start:
                if len(time_ranges) == 0:
                    raise ValueError(
                        "Could not determine temporal start of dataset. "
                        "Please use 'time_range' parameter."
                    )
                temporal_start = time_ranges[0][0]
            if not temporal_end:
                if len(time_ranges) == 0:
                    raise ValueError(
                        "Could not determine temporal end of dataset. "
                        "Please use 'time_range' parameter."
                    )
                temporal_end = time_ranges[-1][1]
        return temporal_start, temporal_end
